﻿using Avalonia.Controls;
using Avalonia.Platform.Storage;
using MsBox.Avalonia;
using MsBox.Avalonia.Enums;
using NPOI.SS.UserModel;
using NPOI.XSSF.UserModel;
using SocialEducatorAvalonia.DTOs;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.IO;
using System.Linq;
using System.Runtime.ExceptionServices;
using System.Threading.Tasks;
using ICell = NPOI.SS.UserModel.ICell;

namespace SocialEducatorAvalonia.Utilities
{
    public static class ExcelUtilities
    {
        #region file manipulation        
        /// <summary>
        /// Проверяет возможно ли открыть файл
        /// </summary>
        /// <param name="filePath"></param>
        /// <returns></returns>
        public static bool CheckIsFileOpened(string filePath)
        {
            try
            {
                using var stream = File.Open(filePath, FileMode.Open, FileAccess.Read, FileShare.None);
                return true;
            }
            catch (Exception)
            {
                MessageBoxManager.GetMessageBoxStandard("Сообщение для вас", "Пожалуйста закройте файл перед тем, как его открыть в программе", ButtonEnum.Ok, Icon.Warning).ShowAsync();
                return false;
                throw;
            }
        }

        /// <summary>
        /// Открытие файла через специальное окно
        /// </summary>
        /// <param name="currentWindow">Окно контекста из которого происходит открывается файл</param>
        /// <returns>Файл найден - file, не найден - null </returns>     
        public static async Task<IReadOnlyList<IStorageFile>>? OpenFileDialogAsync(Window currentWindow)
        {
            // Подготовка вызова окна
            var topLevel = TopLevel.GetTopLevel(currentWindow);

            if (topLevel != null)
            {
                // Настройка FilePickerOpenOptions
                var files = await topLevel.StorageProvider.OpenFilePickerAsync(new FilePickerOpenOptions
                {
                    Title = "Выберите файлы", /*Название окна*/
                    FileTypeFilter = [ new("Excel") { Patterns = ["*.xlsx", "*.xlsb", "*.csv"],
                    MimeTypes = ["*/*"] }, FilePickerFileTypes.All ], /*Настройка для фильтрации файлов*/
                    AllowMultiple = true /*Можно ли выбирать несколько файлов*/
                });

                return files;
            }

            return null;
        }

        public static async Task<IStorageFile>? SaveThatFileAsync(Window currentWindow)
        {
            var topLevel = TopLevel.GetTopLevel(currentWindow);

            if (topLevel != null)
            {
                // Настройка FilePickerOpenOptions
                var file = await topLevel.StorageProvider.SaveFilePickerAsync(new FilePickerSaveOptions
                {
                    Title = "Сохранить файл",
                    FileTypeChoices = [new("Excel") { Patterns = ["*.xlsx", "*.xlsb", "*.csv"], MimeTypes = ["*/*"] }, FilePickerFileTypes.All],
                    DefaultExtension = ".xlsx"
                });

                if (file is not null)
                    return file;
            }

            return null;
        }
        #endregion


        #region для студентов это excel worker
        /// <summary>
        /// Удаление полей из листа
        /// </summary>
        /// <param name="person">передаваемый студент</param>
        /// <param name="persons">Редактируемый список</param>
        public static void DeleteTheField(Persons person, ref ObservableCollection<Persons> persons)
        {
            if (persons.Count == 1)
                persons.Clear();
            else
                persons.Remove(person);
        }

        /// <summary>
        /// Добавление студентов в список
        /// </summary>
        /// <param name="data">информация из excel</param>
        /// <param name="_personsInStorage">лист в который идут новые записи</param>
        public static void AddStudentsFioAndGroup(ref DataSet data, ref List<Persons> _personsInStorage)
        {
            for (int j = 0; j < data.Tables.Count; j++)
            {
                List<Persons> persons = [.. data.Tables[j].AsEnumerable()
                    .Select(dataRow => new Persons
                    {
                        FIO = dataRow.Field<string>("ФИО")!,
                        Group = dataRow.Field<string>("Группа")!
                    })];

                _personsInStorage.AddRange(persons);
            }
        }

        public static void AddPaymentInfo(ref DataSet data, ref List<PersonsPayment> list)
        {
            for (int i = 0; i < data.Tables.Count; i++)
            {
                List<PersonsPayment> payments = [.. data.Tables[i].AsEnumerable().Select(datarow => new PersonsPayment {
                    FIO = datarow.Field<string>("ФИО")!,
                    Group = datarow.Field<string>("Группа")!,
                    Payment = (decimal)datarow.Field<double>("Выплаты")!,
                    PaymentType = datarow.Field<string>("Тип выплаты")!,
                    CurrentPaymentDate = datarow.Field<DateTime>("Начальная дата")!,
                    EndRangeDate = datarow.Field<DateTime>("Конечная дата")!,
                })];

                list.AddRange(payments);
            }
        }


        /// <summary>
        /// Установка Header
        /// </summary>
        /// <param name="sheet"></param>
        /// <param name="columnFIO"></param>
        /// <param name="columnGroup"></param>
        /// <param name="firstRowValues"></param>
        static void SetHeaderValues(ref ISheet sheet, ref List<ICell>? firstRowValues)
        {
            var secondRowValues = sheet.GetRow(1).Cells;

            var rowValues = secondRowValues.Select(x => x.StringCellValue).ToList();

            string? fioCell = rowValues.MaxBy(x => x.Length);

            int columnFIO = secondRowValues.Find(x => x.StringCellValue == fioCell)!.ColumnIndex;

            rowValues.Remove(fioCell!);
            int columnGroup = secondRowValues.Find(x => x.StringCellValue == rowValues.First())!.ColumnIndex;

            firstRowValues[columnFIO].SetCellValue("ФИО");
            firstRowValues[columnGroup].SetCellValue("Группа");
        }


        static void SetHeaderValuesPayments(ref ISheet sheet, ref List<ICell>? firstRowValues)
        {
            var secondRowValues = sheet.GetRow(1).Cells;

            // TODO: возможно стоит доработать но это тяжелая валидация

            List<string?> strings = new();
            List<ICell> stringCells = new();
            List<DateTime> dates = new();
            List<ICell> datesCells = new();
            List<decimal> decimals = new();
            List<ICell> decimalsCells = new();

            foreach (var item in secondRowValues)
            {
                switch (item.CellType)
                {
                    case CellType.Numeric:

                        if (DateUtil.IsCellDateFormatted(item))
                        {
                            DateTime? date = item.DateCellValue;
                            if (date.HasValue)
                            {
                                dates.Add((DateTime)date);
                                datesCells.Add(item);
                            }
                        }
                        else
                        {
                            decimal value = (decimal)item.NumericCellValue;
                            decimals.Add(value);
                            decimalsCells.Add(item);
                        }
                        break;

                    case CellType.String:
                        strings.Add(item.StringCellValue);
                        stringCells.Add(item);
                        break;

                    default:
                        break;
                }
            }

            if (decimals.Count == 1)
                firstRowValues[decimalsCells[0].ColumnIndex].SetCellValue("Выплаты");

            int countWithTwoSpacebar = strings.Count(x => x.Count(c => c == ' ') == 2);

            var groupItem = stringCells.Find(x => x.StringCellValue == strings.MinBy(x => x.Length));
            firstRowValues[groupItem.ColumnIndex].SetCellValue("Группа");
            stringCells.Remove(groupItem);
            strings.Remove(strings.MinBy(_ => _.Length));

            if (countWithTwoSpacebar == 2)
            {
                // два элемента с двумя пробелами

                // в целом ФИО поле больше символов чем выплаты (скорее не верное суждение но хоть как то)

                var cellFio = stringCells.First(x => x.StringCellValue == strings.MaxBy(_ => _.Length));
                firstRowValues[cellFio.ColumnIndex].SetCellValue("ФИО");
                stringCells.Remove(cellFio);
                strings.Remove(strings.First(x => x == strings.MaxBy(_ => _.Length)));

                firstRowValues[stringCells[0].ColumnIndex].SetCellValue("Тип выплаты");
            }
            else
            {
                // нахожу фио через два пробела в строке
                var cellFio = stringCells.First(x => x.StringCellValue == strings.First(x => x.Count(c => c == ' ') == 2));
                firstRowValues[cellFio.ColumnIndex].SetCellValue("ФИО");
                stringCells.Remove(cellFio);
                strings.Remove(strings.First(x => x.Count(c => c == ' ') == 2));

                firstRowValues[stringCells[0].ColumnIndex].SetCellValue("Тип выплаты");
            }


            int result = DateTime.Compare(dates[0], dates[1]);

            if (result < 0)
            {
                // dates[0] раньше dates[1]

                var strDate = datesCells.Find(x => x.DateCellValue == dates[0]);
                firstRowValues[strDate.ColumnIndex].SetCellValue("Начальная дата");
                datesCells.Remove(strDate);

                firstRowValues[datesCells[0].ColumnIndex].SetCellValue("Конечная дата");
            }
            else if (result > 0)
            {
                // dates[0] позже dates[1]

                var strDate = datesCells.Find(x => x.DateCellValue == dates[0]);
                firstRowValues[strDate.ColumnIndex].SetCellValue("Конечная дата");
                datesCells.Remove(strDate);

                firstRowValues[datesCells[0].ColumnIndex].SetCellValue("Начальная дата");
            }
            else
            {
                // даты равны

                var strDate = datesCells.Find(x => x.DateCellValue == dates[0]);
                firstRowValues[strDate.ColumnIndex].SetCellValue("Начальная дата");
                datesCells.Remove(strDate);

                firstRowValues[datesCells[0].ColumnIndex].SetCellValue("Конечная дата");
            }

        }

        public static bool CheckFirstRowPayments(ref IReadOnlyList<IStorageFile> files)
        {
            for (int i = 0; i < files.Count; i++)
            {
                IWorkbook workbook;

                if (!CheckIsFileOpened(files[i].Path.LocalPath))
                    return false;

                // забрали весь файл и теперь его редактируем
                using (FileStream file = new(files[i].Path.LocalPath, FileMode.Open, FileAccess.ReadWrite))
                {
                    workbook = new XSSFWorkbook(file);
                    file.Close();
                }

                for (int j = 0; j < workbook.NumberOfSheets; j++)
                {
                    ISheet sheet = workbook.GetSheetAt(j);
                    IRow firstRow = sheet.GetRow(0);

                    if (firstRow != null)
                    {
                        // если что то есть
                        if (firstRow.Cells.Count <= 6)
                        {
                            while (firstRow.Cells.Count != 6)
                            {
                                firstRow.CreateCell(firstRow.Cells.Count);
                            }
                        }

                        var firstRowValues = sheet.GetRow(0).Cells;

                        var fio = firstRowValues.FirstOrDefault(x => x.StringCellValue == "ФИО");
                        var group = firstRowValues.FirstOrDefault(x => x.StringCellValue == "Группа");
                        var payment = firstRowValues.FirstOrDefault(x => x.StringCellValue == "Выплаты");
                        var typePayment = firstRowValues.FirstOrDefault(x => x.StringCellValue == "Тип выплаты");
                        var strDate = firstRowValues.FirstOrDefault(x => x.StringCellValue == "Начальная дата");
                        var endDate = firstRowValues.FirstOrDefault(x => x.StringCellValue == "Конечная дата");

                        if (fio is null || group is null || payment is null || typePayment is null
                            || strDate is null || endDate is null)
                            SetHeaderValuesPayments(ref sheet, ref firstRowValues);
                    }
                    else
                    {
                        // если ничего нету
                        firstRow = sheet.CreateRow(0);

                        firstRow.CreateCell(0);
                        firstRow.CreateCell(1);
                        firstRow.CreateCell(2);
                        firstRow.CreateCell(3);
                        firstRow.CreateCell(4);
                        firstRow.CreateCell(5);

                        var firstRowValues = sheet.GetRow(0).Cells;

                        SetHeaderValuesPayments(ref sheet, ref firstRowValues);
                    }
                }

                string path = files[i].Path.LocalPath;
                File.Delete(path);

                using (FileStream file = new(files[i].Path.LocalPath, FileMode.OpenOrCreate, FileAccess.Write))
                {
                    workbook.Write(file);
                    file.Close();
                }
            }

            return true;
        }

        /// <summary>
        /// Перепроверка первого поля для студентов
        /// </summary>
        /// <param name="files">Файлы</param>
        public static bool CheckFirstRow(ref IReadOnlyList<IStorageFile> files)
        {
            for (int i = 0; i < files.Count; i++)
            {
                IWorkbook workbook;

                if (!CheckIsFileOpened(files[i].Path.LocalPath))
                {
                    return false;
                }

                // забрали весь файл и теперь его редактируем
                using (FileStream file = new(files[i].Path.LocalPath, FileMode.Open, FileAccess.ReadWrite))
                {
                    workbook = new XSSFWorkbook(file);
                    file.Close();
                }

                // теперь идем по каждому листу
                for (int j = 0; j < workbook.NumberOfSheets; j++)
                {
                    ISheet sheet = workbook.GetSheetAt(j);

                    IRow firstRow = sheet.GetRow(0);

                    if (firstRow is not null)
                    {

                        if (firstRow.Cells.Count <= 2)
                        {
                            while (firstRow.Cells.Count != 2)
                            {
                                firstRow.CreateCell(firstRow.Cells.Count);
                            }
                        }

                        var firstRowValues = sheet.GetRow(0).Cells;

                        var first = firstRowValues.FirstOrDefault(x => x.StringCellValue == "ФИО");
                        var second = firstRowValues.FirstOrDefault(x => x.StringCellValue == "Группа");

                        // Ищу в хедере эти поля
                        if (first is null || second is null)
                            SetHeaderValues(ref sheet, ref firstRowValues);
                    }
                    else
                    {
                        // нужно создать поля хедера

                        firstRow = sheet.CreateRow(0);
                        firstRow.CreateCell(0);
                        firstRow.CreateCell(1);

                        var firstRowValues = sheet.GetRow(0).Cells;
                        SetHeaderValues(ref sheet, ref firstRowValues);
                    }
                }

                string path = files[i].Path.LocalPath;
                File.Delete(path);

                using (FileStream file = new(files[i].Path.LocalPath, FileMode.OpenOrCreate, FileAccess.Write))
                {
                    workbook.Write(file);
                    file.Close();
                }
            }

            return true;
        }
        #endregion
    }
}
